"""
Code to support running in parallel
"""
import logging
import time
from concurrent.futures import as_completed

from ...progress import progress

LOGGER = logging.getLogger(__name__)


def _run_serial(func, configs, config_are_kwargs, desc):
    LOGGER.debug("Entering _run_serial")

    if config_are_kwargs:
        LOGGER.debug("Treating config as kwargs")
        res = [func(**a) for a in progress(configs, desc=desc)]
    else:
        LOGGER.debug("Treating config as args")
        res = [func(a) for a in progress(configs, desc=desc)]

    LOGGER.debug("Exiting _run_serial")
    return res


def _run_parallel(  # pylint:disable=too-many-arguments
    pool, timeout, func, configs, config_are_kwargs, desc, bar_start
):
    LOGGER.debug("Entering _run_parallel")

    if config_are_kwargs:
        LOGGER.debug("Treating config as kwargs")
        futures = [pool.submit(func, **a) for a in configs]
    else:
        LOGGER.debug("Treating config as args")
        futures = [pool.submit(func, a) for a in configs]

    LOGGER.debug("Waiting for jobs to complete")
    for i, future in progress(
        enumerate(as_completed(futures, timeout=timeout)), total=len(futures), desc=desc
    ):
        if future.exception() is not None:
            time.sleep(2)  # let buffer flush out
            print(
                "One of the processes failed, see error below (was something "
                "unable to be pickled?)"
            )
            raise future.exception()

        LOGGER.debug("Job %s completed", i + bar_start)

    res = []

    LOGGER.debug("Collecting results")
    for i, future in enumerate(futures):
        try:
            res.append(future.result())
            LOGGER.debug("Retrived result %s", i + bar_start)
        except Exception as exc:  # pylint:disable=broad-except
            LOGGER.debug("Retrieving result %s failed", i + bar_start)
            res.append(exc)

    LOGGER.debug("Exiting _run_parallel")
    return res


def _parallel_process(  # pylint:disable=too-many-arguments
    func,
    configuration,
    pool=None,
    config_are_kwargs=False,
    front_serial=3,
    front_parallel=2,
    timeout=None,
):
    """
    Run a process in parallel with a progress bar.

    Adapted from http://danshiebler.com/2016-09-14-parallel-progress-bar/

    Parameters
    ----------
    func : function
        A function to apply to each set of arguments in ``configuration``

    configuration : sequence
        An array of configuration with which to run ``func``.

    pool : :obj:`concurrent.futures.ProcessPoolExecutor`
        Pool in which to execute the jobs. If ``None``, the jobs will be executed
        serially in a single process (useful for debugging and benchmarking).

    config_are_kwargs : bool
        Are the elements of ``configuration`` intended to be used as keyword arguments
        when calling ``func``.

    front_serial : int
        The number of iterations to run serially before kicking off the parallel job.
        Useful for debugging.

    front_parallel : int
        The number of initial iterations to run parallel before kicking off the rest
        of the parallel jobs. Useful for debugging (especially if pickling is
        possible).

    timeout : float
        How long to wait for processes to complete before timing out. If
        ``None``, there is no timeout limit.

    Returns
    -------
    sequence
        Results of calling ``func`` with each configuration in ``configuration``
    """
    front_serial_res = []
    if front_serial > 0:
        LOGGER.debug("Running front serial jobs")
        front_serial_res = _run_serial(
            func=func,
            configs=configuration[:front_serial],
            config_are_kwargs=config_are_kwargs,
            desc="Front serial",
        )

    if pool is None:
        LOGGER.info("No pool provided, running rest of the jobs serially and returning")
        rest = _run_serial(
            func=func,
            configs=configuration[front_serial:],
            config_are_kwargs=config_are_kwargs,
            desc="Serial runs",
        )

        return rest + front_serial_res

    front_parallel_res = []
    if front_parallel > 0:
        LOGGER.debug("Running front parallel jobs")
        front_parallel_res = _run_parallel(
            pool=pool,
            timeout=timeout,
            func=func,
            configs=configuration[front_serial : front_serial + front_parallel],
            config_are_kwargs=config_are_kwargs,
            desc="Front parallel",
            bar_start=front_serial,
        )

    LOGGER.debug("Running rest of parallel jobs")
    rest = _run_parallel(
        pool=pool,
        timeout=timeout,
        func=func,
        configs=configuration[front_serial + front_parallel :],
        config_are_kwargs=config_are_kwargs,
        desc="Parallel runs",
        bar_start=front_serial + front_parallel,
    )

    return front_serial_res + front_parallel_res + rest



import os.path

import numpy as np
import numpy.testing as npt
import pytest
from scmdata import ScmRun

import openscm_runner.testing
from openscm_runner import run
from openscm_runner.utils import calculate_quantiles

RTOL = 1e-5


def _check_res(exp, check_val, raise_error, rtol=RTOL):
    try:
        npt.assert_allclose(exp, check_val, rtol=rtol)
    except AssertionError:
        if raise_error:
            raise

        print("exp: {}, check_val: {}".format(exp, check_val))


@pytest.mark.magicc
def test_multimodel_run(test_scenarios, test_data_dir, update_expected_values):
    expected_output_file = os.path.join(
        test_data_dir,
        "expected-integration-output",
        "expected_run_multimodel_output.json",
    )

    res = run(
        climate_models_cfgs={
            "FaIR": [
                {},
                {"q": np.array([0.3, 0.45]), "r0": 30.0, "lambda_global": 0.9},
                {"q": np.array([0.35, 0.4]), "r0": 25.0, "lambda_global": 1.1},
            ],
            "MAGICC7": [
                {
                    "core_climatesensitivity": 3,
                    "rf_soxi_dir_wm2": -0.2,
                    "out_temperature": 1,
                    "out_forcing": 1,
                    "out_dynamic_vars": [
                        "DAT_CO2_CONC",
                        "DAT_AEROSOL_ERF",
                        "DAT_HEATCONTENT_AGGREG_TOTAL",
                        "DAT_HEATUPTK_AGGREG",
                    ],
                    "out_ascii_binary": "BINARY",
                    "out_binary_format": 2,
                },
                {
                    "core_climatesensitivity": 2,
                    "rf_soxi_dir_wm2": -0.1,
                    "out_temperature": 1,
                    "out_forcing": 1,
                    "out_dynamic_vars": [
                        "DAT_CO2_CONC",
                        "DAT_AEROSOL_ERF",
                        "DAT_HEATCONTENT_AGGREG_TOTAL",
                        "DAT_HEATUPTK_AGGREG",
                    ],
                    "out_ascii_binary": "BINARY",
                    "out_binary_format": 2,
                },
                {
                    "core_climatesensitivity": 5,
                    "rf_soxi_dir_wm2": -0.35,
                    "out_temperature": 1,
                    "out_forcing": 1,
                    "out_dynamic_vars": [
                        "DAT_CO2_CONC",
                        "DAT_AEROSOL_ERF",
                        "DAT_HEATCONTENT_AGGREG_TOTAL",
                        "DAT_HEATUPTK_AGGREG",
                    ],
                    "out_ascii_binary": "BINARY",
                    "out_binary_format": 2,
                },
            ],
        },
        scenarios=test_scenarios.filter(scenario=["ssp126", "ssp245", "ssp370"]),
        output_variables=(
            "Surface Air Temperature Change",
            "Atmospheric Concentrations|CO2",
            "Heat Content|Ocean",
            "Heat Content",
            "Heat Uptake|Ocean",
            "Heat Uptake",
            "Effective Radiative Forcing",
            "Effective Radiative Forcing|Aerosols",
            "Effective Radiative Forcing|CO2",
        ),
        out_config=None,
    )

    assert isinstance(res, ScmRun)

    climate_models = res.get_unique_meta("climate_model")
    assert any(["MAGICC" in m for m in climate_models])
    assert any(["FaIR" in m for m in climate_models])

    for climate_model in climate_models:
        res_cm = res.filter(climate_model=climate_model)
        assert set(res_cm.get_unique_meta("variable")) == set(
            [
                "Surface Air Temperature Change",
                "Effective Radiative Forcing",
                "Effective Radiative Forcing|Aerosols",
                "Effective Radiative Forcing|CO2",
                "Heat Content|Ocean",
                "Heat Content",
                "Heat Uptake|Ocean",
                "Heat Uptake",
                "Atmospheric Concentrations|CO2",
            ]
        )

    # check we can also calcluate quantiles
    assert "run_id" in res.meta
    quantiles = calculate_quantiles(res, [0, 0.05, 0.17, 0.5, 0.83, 0.95, 1])
    assert "run_id" not in quantiles.meta

    openscm_runner.testing._check_output(
        res, expected_output_file, rtol=RTOL, update=update_expected_values
    )



import numpy.testing as npt

from openscm_runner.adapters.fair_adapter._scmdf_to_emissions import scmdf_to_emissions


# All emissions that are not covered by FaIR (see
# EMISSIONS_SPECIES_UNITS_CONTEXT) should be ignored, e.g.
# 'Emissions|CO2' (which is the sum of any sub-categories)
def test_emissions_to_ignore(test_scenario_ssp370_world):
    emissions = scmdf_to_emissions(test_scenario_ssp370_world)

    npt.assert_allclose(
        emissions[0, 1], 2.59244179e-03, rtol=1e-5,
    )
    assert emissions.shape[1] == 40



"""
Utility functions for CICERO-SCM adapter
"""


def _get_unique_index_values(idf, index_col, assert_all_same=True):
    """
    Get unique values in index column from a dataframe

    Parameters
    ----------
    idf : :obj:`pd.DataFrame`
        Dataframe to get index values from

    index_col : str
        Column in index to get the values for

    assert_all_same : bool
        Should we assert that all the values are the same before
        returning? If True, only a single value is returned. If
        False, a list is returned.

    Returns
    -------
    str, list
        Values found, either a string or a list depending on
        ``assert_all_same``.

    Raises
    ------
    AssertionError
        ``assert_all_same`` is True and there's more than one
        unique value.
    """
    out = idf.index.get_level_values(index_col).unique().tolist()
    if assert_all_same:
        if len(out) > 1:
            raise AssertionError(out)

        return out[0]

    return out
